
/*
 * Copyright © 2006 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting documentation, and
 * that the name of the copyright holders not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  The copyright holders make no representations
 * about the suitability of this software for any purpose.  It is provided "as
 * is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */
/*
 * This Xinerama implementation comes from the SiS driver which has
 * the following notice:
 */
/* 
 * SiS driver main code
 *
 * Copyright (C) 2001-2005 by Thomas Winischhofer, Vienna, Austria.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1) Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3) The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Author: Thomas Winischhofer <thomas@winischhofer.net>
 *	- driver entirely rewritten since 2001, only basic structure taken from
 *	  old code (except sis_dri.c, sis_shadow.c, sis_accel.c and parts of
 *	  sis_dga.c; these were mostly taken over; sis_dri.c was changed for
 *	  new versions of the DRI layer)
 *
 * This notice covers the entire driver code unless indicated otherwise.
 *
 * Formerly based on code which was
 * 	     Copyright (C) 1998, 1999 by Alan Hourihane, Wigan, England.
 * 	     Written by:
 *           Alan Hourihane <alanh@fairlite.demon.co.uk>,
 *           Mike Chapman <mike@paranoia.com>,
 *           Juanjo Santamarta <santamarta@ctv.es>,
 *           Mitani Hiroshi <hmitani@drl.mei.co.jp>,
 *           David Thomas <davtom@dream.org.uk>.
 */

#include "xorg-config.h"
#include "xf86.h"

#include "xf86Priv.h"
#include "swaprep.h"
#include "extnsionst.h"  	/* required */
#include <X11/extensions/panoramiXproto.h>
#include "dixstruct.h"

#include "xf86Crtc.h"

#include "windowstr.h"

#define xf86_XINERAMA_MAJOR_VERSION   1
#define xf86_XINERAMA_MINOR_VERSION   1

/* Xinerama is not multi-screen capable; just report about screen 0 */
#define xf86_XINERAMA_SCREEN  0

static int Procxf86XineramaQueryVersion(ClientPtr client);
static int Procxf86XineramaGetState(ClientPtr client);
static int Procxf86XineramaGetScreenCount(ClientPtr client);
static int Procxf86XineramaGetScreenSize(ClientPtr client);
static int Procxf86XineramaIsActive(ClientPtr client);
static int Procxf86XineramaQueryScreens(ClientPtr client);
static int SProcxf86XineramaDispatch(ClientPtr client);

/* Proc */

int
Procxf86XineramaQueryVersion(ClientPtr client)
{
    xPanoramiXQueryVersionReply	  rep;
    register int		  n;

    REQUEST_SIZE_MATCH(xPanoramiXQueryVersionReq);
    rep.type = X_Reply;
    rep.length = 0;
    rep.sequenceNumber = client->sequence;
    rep.majorVersion = xf86_XINERAMA_MAJOR_VERSION;
    rep.minorVersion = xf86_XINERAMA_MINOR_VERSION;
    if(client->swapped) {
        swaps(&rep.sequenceNumber, n);
        swapl(&rep.length, n);
        swaps(&rep.majorVersion, n);
        swaps(&rep.minorVersion, n);
    }
    WriteToClient(client, sizeof(xPanoramiXQueryVersionReply), (char *)&rep);
    return (client->noClientException);
}

int
Procxf86XineramaGetState(ClientPtr client)
{
    REQUEST(xPanoramiXGetStateReq);
    WindowPtr			pWin;
    xPanoramiXGetStateReply	rep;
    register int		n;
    Bool			active = FALSE;

    REQUEST_SIZE_MATCH(xPanoramiXGetStateReq);
    pWin = LookupWindow(stuff->window, client);
    if (!pWin) return BadWindow;

    {
	/* XXX do we need more than this? */
	active = TRUE;
    }

    rep.type = X_Reply;
    rep.length = 0;
    rep.sequenceNumber = client->sequence;
    rep.state = active;
    rep.window = stuff->window;
    if(client->swapped) {
       swaps (&rep.sequenceNumber, n);
       swapl (&rep.length, n);
       swapl (&rep.window, n);
    }
    WriteToClient(client, sizeof(xPanoramiXGetStateReply), (char *)&rep);
    return client->noClientException;
}


static int
xf86XineramaScreenCount (ScreenPtr pScreen)
{
    int	i, n;
    ScrnInfoPtr		scrn = xf86Screens[pScreen->myNum];
    xf86CrtcConfigPtr   xf86_config = XF86_CRTC_CONFIG_PTR(scrn);

    n = 0;
    /* okay overlapping hacks are us - this assumes max 2 crtcs
       and we know this limitation on the drivers using this */
    for (i = 0; i < xf86_config->num_crtc; i++) {
	xf86CrtcPtr crtc = xf86_config->crtc[i];
	if (crtc->enabled) {
	    if (i == 1) {
		if ((xf86_config->crtc[0]->x == crtc->x) &&
		    (xf86_config->crtc[0]->y == crtc->y))
		    continue;
	    }
	    n++;
	}
    }
    return n;
}

static Bool
xf86XineramaScreenActive (ScreenPtr pScreen)
{
    return xf86XineramaScreenCount (pScreen) > 0;
}

int
Procxf86XineramaGetScreenCount(ClientPtr client)
{
    REQUEST(xPanoramiXGetScreenCountReq);
    WindowPtr				pWin;
    xPanoramiXGetScreenCountReply	rep;
    register int			n;

    REQUEST_SIZE_MATCH(xPanoramiXGetScreenCountReq);
    pWin = LookupWindow (stuff->window, client);
    if(!pWin)  return BadWindow;
    
    rep.type = X_Reply;
    rep.length = 0;
    rep.sequenceNumber = client->sequence;
    rep.ScreenCount = xf86XineramaScreenCount (pWin->drawable.pScreen);
    rep.window = stuff->window;
    if(client->swapped) {
       swaps(&rep.sequenceNumber, n);
       swapl(&rep.length, n);
       swapl(&rep.window, n);
    }
    WriteToClient(client, sizeof(xPanoramiXGetScreenCountReply), (char *)&rep);
    return client->noClientException;
}

int
Procxf86XineramaGetScreenSize(ClientPtr client)
{
    REQUEST(xPanoramiXGetScreenSizeReq);
    WindowPtr				pWin, pRoot;
    ScreenPtr				pScreen;
    xPanoramiXGetScreenSizeReply	rep;
    register int			n;

    REQUEST_SIZE_MATCH(xPanoramiXGetScreenSizeReq);
    pWin = LookupWindow (stuff->window, client);
    if(!pWin) return BadWindow;

    pScreen = pWin->drawable.pScreen;
    pRoot = WindowTable[pScreen->myNum];
    
    rep.type = X_Reply;
    rep.length = 0;
    rep.sequenceNumber = client->sequence;
    rep.width  = pRoot->drawable.width;
    rep.height = pRoot->drawable.height;
    rep.window = stuff->window;
    rep.screen = stuff->screen;
    if(client->swapped) {
       swaps(&rep.sequenceNumber, n);
       swapl(&rep.length, n);
       swapl(&rep.width, n);
       swapl(&rep.height, n);
       swapl(&rep.window, n);
       swapl(&rep.screen, n);
    }
    WriteToClient(client, sizeof(xPanoramiXGetScreenSizeReply), (char *)&rep);
    return client->noClientException;
}

int
Procxf86XineramaIsActive(ClientPtr client)
{
    xXineramaIsActiveReply	rep;

    REQUEST_SIZE_MATCH(xXineramaIsActiveReq);
	
    rep.type = X_Reply;
    rep.length = 0;
    rep.sequenceNumber = client->sequence;
    rep.state = xf86XineramaScreenActive (screenInfo.screens[xf86_XINERAMA_SCREEN]);
    if(client->swapped) {
	register int n;
	swaps(&rep.sequenceNumber, n);
	swapl(&rep.length, n);
	swapl(&rep.state, n);
    }
    WriteToClient(client, sizeof(xXineramaIsActiveReply), (char *) &rep);
    return client->noClientException;
}

int
Procxf86XineramaQueryScreens(ClientPtr client)
{
    xXineramaQueryScreensReply	rep;
    ScreenPtr	pScreen = screenInfo.screens[xf86_XINERAMA_SCREEN];
    ScrnInfoPtr		scrn = xf86Screens[pScreen->myNum];
    xf86CrtcConfigPtr   xf86_config = XF86_CRTC_CONFIG_PTR(scrn);
    CARD32	number;

    REQUEST_SIZE_MATCH(xXineramaQueryScreensReq);

    rep.type = X_Reply;
    rep.sequenceNumber = client->sequence;
    number = rep.number = xf86XineramaScreenCount (pScreen);
    rep.length = rep.number * sz_XineramaScreenInfo >> 2;
    if(client->swapped) {
	register int n;
	swaps(&rep.sequenceNumber, n);
	swapl(&rep.length, n);
	swapl(&rep.number, n);
    }
    WriteToClient(client, sizeof(xXineramaQueryScreensReply), (char *)&rep);

    if(number) {
	xXineramaScreenInfo scratch;
	int i;

	for(i = 0; i < number; i++) {
	    xf86CrtcPtr	crtc = xf86_config->crtc[i];
	    if (crtc->enabled)
	    {
	        int width, height;
                width = xf86ModeWidth(&crtc->mode, crtc->rotation);
                height = xf86ModeHeight(&crtc->mode, crtc->rotation);
		scratch.x_org  = crtc->x;
		scratch.y_org  = crtc->y;
		scratch.width  = width;
		scratch.height = height;
		if(client->swapped) {
		    register int n;
		    swaps(&scratch.x_org, n);
		    swaps(&scratch.y_org, n);
		    swaps(&scratch.width, n);
		    swaps(&scratch.height, n);
		}
		WriteToClient(client, sz_XineramaScreenInfo, (char *)&scratch);
	    }
	}
    }

    return client->noClientException;
}



static int
Procxf86XineramaDispatch(ClientPtr client)
{
    REQUEST(xReq);
    switch (stuff->data) {
	case X_PanoramiXQueryVersion:
	     return Procxf86XineramaQueryVersion(client);
	case X_PanoramiXGetState:
	     return Procxf86XineramaGetState(client);
	case X_PanoramiXGetScreenCount:
	     return Procxf86XineramaGetScreenCount(client);
	case X_PanoramiXGetScreenSize:
	     return Procxf86XineramaGetScreenSize(client);
	case X_XineramaIsActive:
	     return Procxf86XineramaIsActive(client);
	case X_XineramaQueryScreens:
	     return Procxf86XineramaQueryScreens(client);
    }
    return BadRequest;
}

/* SProc */

static int
SProcxf86XineramaQueryVersion (ClientPtr client)
{
    REQUEST(xPanoramiXQueryVersionReq);
    register int n;
    swaps(&stuff->length,n);
    REQUEST_SIZE_MATCH (xPanoramiXQueryVersionReq);
    return Procxf86XineramaQueryVersion(client);
}

static int
SProcxf86XineramaGetState(ClientPtr client)
{
    REQUEST(xPanoramiXGetStateReq);
    register int n;
    swaps (&stuff->length, n);
    REQUEST_SIZE_MATCH(xPanoramiXGetStateReq);
    swapl (&stuff->window, n);
    return Procxf86XineramaGetState(client);
}

static int
SProcxf86XineramaGetScreenCount(ClientPtr client)
{
    REQUEST(xPanoramiXGetScreenCountReq);
    register int n;
    swaps (&stuff->length, n);
    REQUEST_SIZE_MATCH(xPanoramiXGetScreenCountReq);
    swapl (&stuff->window, n);
    return Procxf86XineramaGetScreenCount(client);
}

static int
SProcxf86XineramaGetScreenSize(ClientPtr client)
{
    REQUEST(xPanoramiXGetScreenSizeReq);
    register int n;
    swaps (&stuff->length, n);
    REQUEST_SIZE_MATCH(xPanoramiXGetScreenSizeReq);
    swapl (&stuff->window, n);
    swapl (&stuff->screen, n);
    return Procxf86XineramaGetScreenSize(client);
}

static int
SProcxf86XineramaIsActive(ClientPtr client)
{
    REQUEST(xXineramaIsActiveReq);
    register int n;
    swaps (&stuff->length, n);
    REQUEST_SIZE_MATCH(xXineramaIsActiveReq);
    return Procxf86XineramaIsActive(client);
}

static int
SProcxf86XineramaQueryScreens(ClientPtr client)
{
    REQUEST(xXineramaQueryScreensReq);
    register int n;
    swaps (&stuff->length, n);
    REQUEST_SIZE_MATCH(xXineramaQueryScreensReq);
    return Procxf86XineramaQueryScreens(client);
}

static int
SProcxf86XineramaDispatch(ClientPtr client)
{
    REQUEST(xReq);
    switch (stuff->data) {
	case X_PanoramiXQueryVersion:
	     return SProcxf86XineramaQueryVersion(client);
	case X_PanoramiXGetState:
	     return SProcxf86XineramaGetState(client);
	case X_PanoramiXGetScreenCount:
	     return SProcxf86XineramaGetScreenCount(client);
	case X_PanoramiXGetScreenSize:
	     return SProcxf86XineramaGetScreenSize(client);
	case X_XineramaIsActive:
	     return SProcxf86XineramaIsActive(client);
	case X_XineramaQueryScreens:
	     return SProcxf86XineramaQueryScreens(client);
    }
    return BadRequest;
}

static void
xf86XineramaResetProc(ExtensionEntry* extEntry)
{
}

void xf86XineramaExtensionInit(void)
{
#ifdef PANORAMIX
    if(!noPanoramiXExtension)
	return;
#endif

    /*
     * Xinerama isn't capable enough to have multiple protocol screens each
     * with their own output geometry.  So if there's more than one protocol
     * screen, just don't even try.
     */
    if (screenInfo.numScreens > 1)
	return;

    (void) AddExtension(PANORAMIX_PROTOCOL_NAME, 0,0,
			Procxf86XineramaDispatch,
			SProcxf86XineramaDispatch,
			xf86XineramaResetProc,
			StandardMinorOpcode);
}
